@@ -1,5 +1,5 @@
-from .filters import all_filters, filtdict
-from .models import blackbody_to_filters, planck_fast
+from .filters import all_filters, filtdict, extinction_law
+from .models import planck_fast, UniformPrior, LogUniformPrior, GaussianPrior
 from .lightcurve import LC
 
 from astropy import constants as const
@@ -52,20 +52,56 @@ def pseudo(temp, radius, z, filter0=filtdict['I'], filter1=filtdict['U'], cutoff
     return L_opt
 
 
-def blackbody_mcmc(epoch1, z, p0=None, show=False, outpath='.', nwalkers=10, burnin_steps=200, steps=100,
-                   T_range=(1., 100.), R_range=(0.01, 1000.), cutoff_freq=np.inf, prev_temp=None, save_chains=False,
-                   use_sigma=False, sigma_type='relative', sigma_max=10.):
+def plot_chain(chain, labels=None):
     """
-    Fit a blackbody spectrum to a spectral energy distribution using a Markov-chain Monte Carlo routine
+    Plot the chain histories for an MCMC run
 
     Parameters
     ----------
+    chain : array-like
+        Array containing the MCMC chain history, e.g., from :meth:`emcee.EnsembleSampler.chain`
+    labels : iterable, optional
+        A list of axis labels for each parameter in the chain
+
+    Returns
+    -------
+    fig : matplotlib.pyplot.Figure
+        Figure object containing the chain history plots
+    """
+    ndim = chain.shape[-1]
+    fig, ax = plt.subplots(ndim, figsize=(6., 2. * ndim))
+    for i in range(ndim):
+        ax[i].plot(chain[:, :, i].T, 'k', alpha=0.2)
+        if labels:
+            ax[i].set_ylabel(labels[i])
+    return fig
+
+
+def spectrum_mcmc(spectrum, epoch1, priors, starting_guesses, z=0., ebv=0., spectrum_kwargs=None, show=False, outpath='.',
+                  nwalkers=10, burnin_steps=200, steps=100, save_chains=False, use_sigma=False, sigma_type='relative',
+                  labels=None, freq_min=100., freq_max=1000.):
+    """
+    Fit the given spectral energy distribution to an epoch of photometry using a Markov-chain Monte Carlo routine
+
+    Parameters
+    ----------
+    spectrum : function
+        Function describing the spectrum. The first argument must be frequency in THz, and it must return spectral
+        luminosity in watts per hertz. For a blackbody, you can use :func:`.models.planck_fast`.
     epoch1 : lightcurve_fitting.lightcurve.LC
         A single "epoch" of photometry that defines the observed spectral energy distribution
-    z : float
-        Redshift between the blackbody (rest frame) and the filters (observed frame)
-    p0 : list, tuple, array-like, optional
-        Initial guess for [temperature (kK), radius (1000 Rsun)]. Default: ``[10., 10.]``
+    priors : list
+        Prior probability distributions for each model parameter (and sigma, if used). Available priors:
+        :class:`.models.UniformPrior` (default), :class:`.models.LogUniformPrior`, :class:`.models.GaussianPrior`
+    starting_guesses : array-like
+        Initial guesses for each input parameter to ``spectrum``. Must have shape (nwalkers, nparameters).
+    z : float, optional
+        Redshift between the emission source and the observed filter. Default: 0.
+    ebv : float, array-like, optional
+        Selective extinction :math:`E(B-V)` in magnitudes, evaluated using a Fitzpatrick (1999) extinction law with
+        :math:`R_V=3.1`. Its shape must be broadcastable to any array-like arguments. Default: 0.
+    spectrum_kwargs : dict, optional
+        Keyword arguments to be passed to the ``spectrum`` function.
     show : bool, optional
         Plot the chain histories, and display all plots at the end of the script. Default: only save the corner plot
     outpath : str, optional
@@ -75,15 +111,7 @@ def blackbody_mcmc(epoch1, z, p0=None, show=False, outpath='.', nwalkers=10, bur
     burnin_steps : int, optional
         Number of MCMC steps before convergence. This part of the history is discarded. Default: 200
     steps : int, optional
-        Number of MCMC steps after convergence. This part of the history is used to calculate paramers. Default: 100.
-    T_range : tuple, list, array-like, optional
-        Range of allowed temperatures (in kilokelvins) in the prior. Default: ``(1., 100.)``
-    R_range : tuple, list, array-like, optional
-        Range of allowed radii (in 1000 solar radii) in the prior. Default: ``(0.01, 1000.)``
-    cutoff_freq : float, optional
-        Cutoff frequency (in terahertz) for a modified blackbody spectrum (see https://doi.org/10.3847/1538-4357/aa9334)
-    prev_temp : scipy.stats.gaussian_kde, optional
-        Temperature distribution from fitting the previous epoch's SED. If given, this is used as the prior.
+        Number of MCMC steps after convergence. This part of the history is used to calculate parameters. Default: 100.
     save_chains : bool, optional
         If True, save the MCMC chain histories to the directory ``outpath``. Default: only save the corner plot
     use_sigma : bool, optional
@@ -91,8 +119,10 @@ def blackbody_mcmc(epoch1, z, p0=None, show=False, outpath='.', nwalkers=10, bur
     sigma_type : str, optional
         If 'relative' (default), sigma will be in units of the individual photometric uncertainties.
         If 'absolute', sigma will be in units of the median photometric uncertainty.
-    sigma_max : float, optional
-        Maximum allowed intrinsic scatters (in units of the median uncertainty) in the prior. Default: 10.
+    labels : list, optional
+        Axis labels for the chain histories and corner plot.
+    freq_min, freq_max : float, optional
+        Minimum and maximum frequencies for the SED panel in the corner plot. Default: (100., 1000.) or observed range.
 
     Returns
     -------
@@ -104,18 +134,8 @@ def blackbody_mcmc(epoch1, z, p0=None, show=False, outpath='.', nwalkers=10, bur
     filtobj = epoch1['filter'].data
     mjdavg = np.mean(epoch1['MJD'].data)
 
-    if p0 is None:
-        p0 = [10., 10.]
-
-    def log_prior(p):  # p = [T, R]
-        if (np.any(p[0] < T_range[0]) or np.any(p[0] > T_range[1])
-                or np.any(p[1] < R_range[0]) or np.any(p[1] > R_range[1])
-                or (use_sigma and np.any(p[2] > sigma_max))):
-            return -np.inf
-        elif prev_temp is not None:
-            return prev_temp.logpdf(p[0]) - np.log(p[1]) - (p[2] ** 2. / 2. if use_sigma else 0.)
-        else:
-            return -np.log(p[1]) - (p[2] ** 2. / 2. if use_sigma else 0.)
+    if spectrum_kwargs is None:
+        spectrum_kwargs = {}
 
     if sigma_type == 'relative':
         sigma_units = dy
@@ -124,65 +144,37 @@ def blackbody_mcmc(epoch1, z, p0=None, show=False, outpath='.', nwalkers=10, bur
     else:
         raise Exception('sigma_type must either be "relative" or "absolute"')
 
-    def log_likelihood(p, filtobj, y, dy):
-        y_fit = blackbody_to_filters(filtobj, p[0], p[1], z, cutoff_freq)
-        sigma = np.sqrt(dy ** 2. + (p[2] * sigma_units) ** 2.) if use_sigma else dy
-        return -0.5 * np.sum(np.log(2 * np.pi * sigma ** 2.) + ((y - y_fit) / sigma) ** 2.)
+    def log_posterior(p):
+        log_prior = 0.
+        for prior, p_i in zip(priors, p):
+            log_prior += prior(p_i)
+        if np.isinf(log_prior):
+            return log_prior
+        y_fit = np.array([f.synthesize(spectrum, *p[:-1 if use_sigma else None], z=z, ebv=ebv, **spectrum_kwargs)
+                          for f in filtobj])
+        sigma = np.sqrt(dy ** 2. + (p[-1] * sigma_units) ** 2.) if use_sigma else dy
+        log_likelihood = -0.5 * np.sum(np.log(2 * np.pi * sigma ** 2.) + ((y - y_fit) / sigma) ** 2.)
+        return log_prior + log_likelihood
 
-    def log_posterior(p, filtobj, y, dy):
-        return log_prior(p) + log_likelihood(p, filtobj, y, dy)
-
-    ndim = 2
-    starting_guesses = np.random.randn(nwalkers, ndim) + p0
-    starting_guesses[starting_guesses <= 0.] = 1.
-    if use_sigma:
-        ndim += 1
-        starting_guesses = np.append(starting_guesses, np.abs(np.random.randn(nwalkers, 1)), axis=1)
-
-    sampler = emcee.EnsembleSampler(nwalkers, ndim, log_posterior, args=[filtobj, y, dy])
+    ndim = len(priors)
+    sampler = emcee.EnsembleSampler(nwalkers, ndim, log_posterior)
     pos, _, _ = sampler.run_mcmc(starting_guesses, burnin_steps)
 
     # Plotting
     if show:
-        f1, ax2 = plt.subplots(ndim)
-        for i in range(len(ax2)):
-            ax2[i].plot(sampler.chain[:, :, i].T, 'k', alpha=0.2)
+        plot_chain(sampler.chain, labels)
     sampler.reset()
     sampler.run_mcmc(pos, steps)
     if show:
-        f2, ax3 = plt.subplots(ndim)
-        for i in range(len(ax3)):
-            ax3[i].plot(sampler.chain[:, :, i].T, 'k', alpha=0.2)
-
-    labels = ['T (kK)', 'R (1000 R$_\\odot$)']
-    if use_sigma:
-        labels.append('$\\sigma$')
-    f4 = corner.corner(sampler.flatchain, labels=labels)
-    ax = f4.get_axes()[1]
-    ps = sampler.flatchain[np.random.choice(sampler.flatchain.shape[0], 100)].T
-    xfit = np.arange(100., max(1000., min(filtobj).freq_eff.value))
-    yfit = planck_fast(xfit * (1. + z), ps[0], ps[1], cutoff_freq)
-    plt.sca(ax)
-    epoch1.plot(xcol='freq', ycol='lum', offset_factor=0.)
-    ax.plot(xfit, yfit.T, color='k', alpha=0.05)
-    ax.set_frame_on(True)
-    ax.xaxis.set_major_locator(AutoLocator())
-    ax.xaxis.tick_top()
-    ax.set_xlabel('Frequency (THz)')
-    ax.xaxis.set_label_position('top')
-    ax.yaxis.set_major_locator(AutoLocator())
-    ax.yaxis.tick_right()
-    ax.set_ylabel('$L_\\nu$ (W Hz$^{-1}$)')
-    ax.yaxis.set_label_position('right')
-    f4.tight_layout()
+        plot_chain(sampler.chain, labels)
 
     os.makedirs(outpath, exist_ok=True)
-    filename = os.path.join(outpath, f'{mjdavg:.3f}.png')
-    print(filename)
-    f4.savefig(filename)
     if save_chains:
         chain_filename = os.path.join(outpath, f'{mjdavg:.3f}.npy')
         np.save(chain_filename, sampler.flatchain)
+
+    f4 = spectrum_corner(spectrum, epoch1, sampler.flatchain, z, ebv, spectrum_kwargs, use_sigma, labels,
+                         freq_min=freq_min, freq_max=freq_max, save_plot_as=os.path.join(outpath, f'{mjdavg:.3f}.pdf'))
     if show:
         plt.show()
     else:
@@ -191,6 +183,103 @@ def blackbody_mcmc(epoch1, z, p0=None, show=False, outpath='.', nwalkers=10, bur
     return sampler
 
 
+def spectrum_corner(spectrum, epoch1, sampler_flatchain, z=0., ebv=0., spectrum_kwargs=None, use_sigma=False,
+                    labels=None, freq_min=100., freq_max=1000., save_plot_as=''):
+    """
+    Plot the posterior distributions in a corner (pair) plot, with an inset showing the observed and model SEDs.
+
+    Parameters
+    ----------
+    spectrum : function
+        Function describing the spectrum. The first argument must be frequency in THz, and it must return spectral
+        luminosity in watts per hertz. For a blackbody, you can use :func:`.models.planck_fast`.
+    epoch1 : lightcurve_fitting.lightcurve.LC
+        A single "epoch" of photometry that defines the observed spectral energy distribution
+    sampler_flatchain : array-like
+        2D array containing the aggregated MCMC chain histories
+    z : float, optional
+        Redshift between the emission source and the observed filter. Default: 0.
+    ebv : float, array-like, optional
+        Selective extinction :math:`E(B-V)` in magnitudes, evaluated using a Fitzpatrick (1999) extinction law with
+        :math:`R_V=3.1`. Its shape must be broadcastable to any array-like arguments. Default: 0.
+    spectrum_kwargs : dict, optional
+        Keyword arguments to be passed to the ``spectrum`` function.
+    use_sigma : bool, optional
+        Include an intrinsic scatter parameter. Default: False.
+    labels : list, optional
+        Axis labels for the chain histories and corner plot.
+    freq_min, freq_max : float, optional
+        Minimum and maximum frequencies for the SED panel in the corner plot. Default: (100., 1000.) or observed range.
+    save_plot_as : str, optional
+        Filename to which to save the resulting plot
+
+    Returns
+    -------
+    f4 : matplotlib.pyplot.Figure
+        Figure containing the corner and SED plots.
+    """
+
+    ndim = sampler_flatchain.shape[-1]
+    if spectrum_kwargs is None:
+        spectrum_kwargs = {}
+
+    f4 = corner.corner(sampler_flatchain, labels=labels)
+
+    ps = sampler_flatchain[np.random.choice(sampler_flatchain.shape[0], 100)].T
+    xmin = min(freq_min, epoch1['filter'].max().freq_eff.value)
+    xmax = max(freq_max, epoch1['filter'].min().freq_eff.value)
+    xfit = np.arange(xmin, xmax)
+    freq = xfit * (1. + z)
+    yfit = spectrum(freq, *ps[:-1 if use_sigma else None], **spectrum_kwargs) * extinction_law(freq, ebv)
+    yscale = 10. ** np.floor(np.log10(yfit.max()))
+
+    # add tick labels and axis labels to the top right figure so tight_layout works
+    all_axes = np.reshape(f4.get_axes(), (ndim, ndim))
+    top_right = all_axes[0, -1]
+    top_right.set_frame_on(True)
+    top_right.xaxis.set_major_locator(plt.AutoLocator())
+    top_right.xaxis.tick_top()
+    top_right.set_xlabel('Frequency (THz)')
+    top_right.xaxis.set_label_position('top')
+    top_right.yaxis.set_major_locator(plt.AutoLocator())
+    top_right.yaxis.tick_right()
+    top_right.set_ylabel(f'Luminosity $L_\\nu$ (10$^{{{np.log10(yscale):.0f}}}$ W Hz$^{{-1}}$)')
+    top_right.yaxis.set_label_position('right')
+    f4.tight_layout(h_pad=0.05, w_pad=0.05)
+
+    bottom_left = all_axes[ndim // 2 - 1, (ndim + 1) // 2]
+    if bottom_left is top_right:
+        ax = top_right
+    else:
+        # add a new axis that takes up the entire top right of the figure
+        bbox0 = bottom_left.bbox.transformed(f4.transFigure.inverted())
+        bbox1 = top_right.bbox.transformed(f4.transFigure.inverted())
+        ax = f4.add_axes([bbox0.xmin, bbox0.ymin, bbox1.xmax - bbox0.xmin, bbox1.ymax - bbox0.ymin])
+        ax.xaxis.tick_top()
+        ax.set_xlabel('Frequency (THz)')
+        ax.xaxis.set_label_position('top')
+        ax.yaxis.tick_right()
+        ax.set_ylabel(f'Luminosity $L_\\nu$ (10$^{{{np.log10(yscale):.0f}}}$ W Hz$^{{-1}}$)')
+        ax.yaxis.set_label_position('right')
+
+        # turn the original top right axis back off
+        top_right.set_frame_on(False)
+        top_right.xaxis.set_major_locator(plt.NullLocator())
+        top_right.set_xlabel('')
+        top_right.yaxis.set_major_locator(plt.NullLocator())
+        top_right.set_ylabel('')
+
+    for row in epoch1:
+        ax.errorbar(row['freq'], row['lum'] / yscale, row['dlum'] / yscale, marker='o', **row['filter'].plotstyle)
+    ax.plot(xfit, yfit.T / yscale, color='k', alpha=0.05)
+
+    if save_plot_as:
+        f4.savefig(save_plot_as)
+        print('saving figure as ' + save_plot_as)
+
+    return f4
+
+
 def plot_bolometric_results(t0, save_plot_as=None):
     """
     Plot the parameters and bolometric light curves that result from fitting the spectral energy distribution
@@ -450,12 +539,12 @@ def calc_colors(epoch1, colors):
 
 def plot_color_curves(t, colors=None, fmt='o', limit_length=0.1, xcol='MJD'):
     """
-    Plot the color curves calculated by :func:`calculate_bolometric`.
+    Plot the color curves calculated by :func:`.calculate_bolometric`.
 
     Parameters
     ----------
     t : lightcurve_fitting.lightcurve.LC
-        Output table from :func:`calculate_bolometric`
+        Output table from :func:`.calculate_bolometric`
     colors : list, optional
         List of colors to plot, e.g., ``['g-r', 'r-i']``. By default, plot all recognizable colors.
     fmt : str, optional
@@ -488,10 +577,9 @@ def plot_color_curves(t, colors=None, fmt='o', limit_length=0.1, xcol='MJD'):
     return fig
 
 
-def calculate_bolometric(lc, z, outpath='.', res=1., nwalkers=10, burnin_steps=200, steps=100,
-                         T_range=(1., 100.), R_range=(0.01, 1000.), save_table_as=None, min_nfilt=3,
-                         cutoff_freq=np.inf, show=False, colors=None, do_mcmc=True, save_chains=False, use_sigma=False,
-                         also_group_by=()):
+def calculate_bolometric(lc, z=0., outpath='.', res=1., nwalkers=10, burnin_steps=200, steps=100, priors=None,
+                         save_table_as=None, min_nfilt=3, cutoff_freq=np.inf, show=False, colors=None, do_mcmc=True,
+                         save_chains=False, use_sigma=False, sigma_type='relative', also_group_by=()):
     """
     Calculate the full bolometric light curve from a table of broadband photometry
 
@@ -499,8 +587,8 @@ def calculate_bolometric(lc, z, outpath='.', res=1., nwalkers=10, burnin_steps=2
     ----------
     lc : lightcurve_fitting.lightcurve.LC
         Table of broadband photometry including columns "MJD", "mag", "dmag", "filt"
-    z : float
-        Redshift between the blackbody (rest frame) and the filters (observed frame)
+    z : float, optional
+        Redshift between the emission source and the observed filter. Default: 0.
     outpath : str, optional
         Directory to which to save the corner plots and MCMC chains. Default: current directory
     res : float, optional
@@ -511,10 +599,10 @@ def calculate_bolometric(lc, z, outpath='.', res=1., nwalkers=10, burnin_steps=2
         Number of MCMC steps before convergence. This part of the history is discarded. Default: 200
     steps : int, optional
         Number of MCMC steps after convergence. This part of the history is used to calculate paramers. Default: 100.
-    T_range : tuple, list, array-like, optional
-        Range of allowed temperatures (in kilokelvins) in the prior. Default: ``(1., 100.)``
-    R_range : tuple, list, array-like, optional
-        Range of allowed radii (in 1000 solar radii) in the prior. Default: ``(0.01, 1000.)``
+    priors : list, optional
+        Prior probability distributions for temperature (in kilokelvins) and radius (in 1000 solar radii). Available
+        priors: :class:`.models.UniformPrior`, :class:`.models.LogUniformPrior`, :class:`.models.GaussianPrior`.
+        Default: ``T = Uniform(1., 100.)``, ``R = LogUniform(0.01, 1000.)``, ``σ = Gaussian(0., 10., stddev=1.)``.
     save_table_as : str, optional
         Filename to which to save the output table of blackbody parameters and bolometric luminosities
     min_nfilt : int, optional
@@ -532,6 +620,9 @@ def calculate_bolometric(lc, z, outpath='.', res=1., nwalkers=10, burnin_steps=2
         If True, save the MCMC chain histories to the directory ``outpath``. Default: only save the corner plot
     use_sigma : bool, optional
         Include an intrinsic scatter parameter in the MCMC fit. Default: False.
+    sigma_type : str, optional
+        If 'relative' (default), sigma will be in units of the individual photometric uncertainties.
+        If 'absolute', sigma will be in units of the median photometric uncertainty.
     also_group_by : list, optional
         Group by these columns in addition to epoch
 
@@ -560,6 +651,11 @@ def calculate_bolometric(lc, z, outpath='.', res=1., nwalkers=10, burnin_steps=2
             + [float] * 2 * len(colors) + [bool] * 2 * len(colors) + ['S6'] + ([lc['source'].dtype] if use_src else []),
             masked=True)
 
+    if priors is None:
+        priors = [UniformPrior(1., 100.), LogUniformPrior(0.01, 1000.)]
+        if use_sigma:
+            priors.append(GaussianPrior(0., 10.))
+
     sampler = None
     lc = lc[np.isfinite(lc['dmag']) & (lc['dmag'] > 0.)]
     for epoch1 in group_by_epoch(lc, res, also_group_by):
@@ -582,10 +678,9 @@ def calculate_bolometric(lc, z, outpath='.', res=1., nwalkers=10, burnin_steps=2
             continue
 
         if nfilt > 1:
-            prev_temp = None
-            p0 = [10., 10.]
+            p0 = np.array([10., 10.])
         elif sampler is not None:
-            prev_temp = gaussian_kde(sampler.flatchain[:, 0])
+            priors[0] = gaussian_kde(sampler.flatchain[:, 0]).logpdf
             p0 = np.median(sampler.flatchain, axis=0)
         else:
             continue
@@ -594,19 +689,31 @@ def calculate_bolometric(lc, z, outpath='.', res=1., nwalkers=10, burnin_steps=2
         filtstr = ''.join([f.char for f in sorted(filts)])
 
         # blackbody - least squares
+        T_range = (priors[0].p_min, priors[0].p_max)
+        R_range = (priors[1].p_min, priors[1].p_max)
         try:
             temp, radius, dtemp, drad, lum, dlum, L_opt = blackbody_lstsq(epoch1, z, p0, T_range, R_range, cutoff_freq)
-            p0 = [temp, radius]
+            p0 = np.array([temp, radius])
         except RuntimeError:  # optimization failed
             temp = radius = dtemp = drad = lum = dlum = L_opt = np.nan
 
+        rng = np.random.default_rng()
+        starting_guesses = rng.normal(size=(nwalkers, 2)) + p0
+        starting_guesses[starting_guesses <= 0.] = 1.
+        labels = ['T (kK)', 'R (1000 R$_\\odot$)']
+        if use_sigma:
+            starting_guesses = np.append(starting_guesses, np.abs(rng.normal(size=(nwalkers, 1))), axis=1)
+            labels.append('$\\sigma$')
+
         # blackbody - MCMC
         try:
             if not do_mcmc:
                 raise ValueError
-            sampler = blackbody_mcmc(epoch1, z, p0, outpath=outpath, nwalkers=nwalkers, burnin_steps=burnin_steps,
-                                     steps=steps, T_range=T_range, R_range=R_range, cutoff_freq=cutoff_freq, show=show,
-                                     prev_temp=prev_temp, save_chains=save_chains, use_sigma=use_sigma)
+            spectrum_kwargs = {'cutoff_freq': cutoff_freq}
+            sampler = spectrum_mcmc(planck_fast, epoch1, priors, starting_guesses, z=z, spectrum_kwargs=spectrum_kwargs,
+                                    outpath=outpath, nwalkers=nwalkers, burnin_steps=burnin_steps, steps=steps,
+                                    show=show, save_chains=save_chains, use_sigma=use_sigma, sigma_type=sigma_type,
+                                    labels=labels)
             L_mcmc_opt = pseudo(sampler.flatchain[:, 0], sampler.flatchain[:, 1], z, cutoff_freq=cutoff_freq)
             (T_mcmc, R_mcmc), (dT0_mcmc, dR0_mcmc), (dT1_mcmc, dR1_mcmc) = median_and_unc(sampler.flatchain[:, :2])
             L_mcmc, dL_mcmc0, dL_mcmc1 = median_and_unc(L_mcmc_opt)
