@@ -4,10 +4,33 @@ from astropy.table import Table
 import astropy.units as u
 import astropy.constants as const
 import os
-from .models import planck_fast
 from pkg_resources import resource_filename
 from functools import total_ordering
-from extinction import fitzpatrick99 as extinction_law
+from extinction import fitzpatrick99
+
+c = const.c.to(u.angstrom * u.THz).value
+
+
+def extinction_law(freq, ebv, rv=3.1):
+    """
+    A vectorized version of the Fitzpatrick (1999) extinction law from the ``extinction`` package
+
+    Parameters
+    ----------
+    freq : array-like
+        Frequencies in THz in the frame of the dust
+    ebv : array-like
+        Selective extinction :math:`E(B-V)` in magnitudes
+    rv : float, optional
+        Ratio of total to selective extinction :math:`R_V`. Default: 3.1.
+
+    Returns
+    -------
+    extinction : array-like
+        Extinction factor :math:`10^{A/-2.5}` at each input wavelength.
+    """
+    A = np.squeeze([fitzpatrick99(c / freq, rv * e) for e in np.atleast_1d(ebv)])
+    return 10. ** (A / -2.5)
 
 
 @total_ordering
@@ -117,7 +140,7 @@ class Filter:
             self.textcolor = self.linecolor
         self.italics = italics
         self.system = system
-        if self.system == 'Johnson':
+        if self.system in ['Johnson', 'JWST MIRI']:
             self.mec = 'k'
         else:
             self.mec = self.linecolor
@@ -207,33 +230,28 @@ class Filter:
             self.dfreq = -dfreq
             self.freq_range = (freq_eff.value - freq0, freq1 - freq_eff.value)
 
-    def blackbody(self, T, R, z=0., cutoff_freq=np.inf, ebv=0.):
+    def synthesize(self, spectrum, *args, z=0., ebv=0., **kwargs):
         """
-        Returns the average Lnu of a blackbody in this filter
+        Returns the average Lnu of the given spectrum in this filter
 
         Parameters
         ----------
-        T : float or array-like
-            Temperature of the blackbody in kilokelvins
-        R : float or array-like
-            Radius of the blackbody in thousands of solar radii
+        spectrum : function
+            Function describing the spectrum. The first argument must be frequency in THz, and it must return spectral
+            luminosity in watts per hertz. All arguments are passed to this function except for keywords z and ebv.
         z : float, optional
-            Redshift between the blackbody source and the filter
-        cutoff_freq : float, optional
-            Cutoff frequency of the blackbody in terahertz as defined in https://doi.org/10.3847/1538-4357/aa9334.
-            Default: unmodified blackbody.
+            Redshift between the emission source and the observed filter. Default: 0.
         ebv : float, array-like, optional
-            Selective extinction E(B-V) in magnitudes, evaluated using a Fitzpatrick (1999) extinction law with R_V=3.1.
-            Its shape must be broadcastable to T and R. Default: 0.
+            Selective extinction :math:`E(B-V)` in magnitudes, evaluated using a Fitzpatrick (1999) extinction law with
+            :math:`R_V=3.1`. Its shape must be broadcastable to any array-like arguments. Default: 0.
 
         Returns
         -------
         Lnu : float or array-like
             Average spectral luminosity in the filter in watts per hertz
         """
-        wl = self.trans['wl'].to(u.angstrom).value / (1. + z)
-        A = np.squeeze([extinction_law(wl, 3.1 * e) for e in np.atleast_1d(ebv)])
-        return np.trapz(planck_fast(self.trans['freq'].data * (1. + z), T, R, cutoff_freq) * 10. ** (A / -2.5)
+        freq = self.trans['freq'].value * (1. + z)
+        return np.trapz(spectrum(freq, *args, **kwargs) * extinction_law(freq, ebv)
                         * self.trans['T_norm_per_freq'].data, self.trans['freq'].data)
 
     def spectrum(self, freq, lum, z=0., ebv=0.):
@@ -245,15 +263,15 @@ class Filter:
 
         Parameters
         ----------
-        freq : astropy.units.Quantity
-            Frequency of the input spectrum (with units)
+        freq : float or array-like
+            Frequency of the input spectrum in THz
         lum : float or array-like
             Spectral luminosity (Lnu) or flux (Fnu) of the input spectrum
         z : float, optional
-            Redshift between the spectrum and the filter
+            Redshift between the spectrum and the filter. Default: 0.
         ebv : float, array-like, optional
-            Selective extinction E(B-V) in magnitudes, evaluated using a Fitzpatrick (1999) extinction law with R_V=3.1.
-            Its shape must be broadcastable to T and R. Default: 0.
+            Selective extinction :math:`E(B-V)` in magnitudes, evaluated using a Fitzpatrick (1999) extinction law with
+            :math:`R_V=3.1`. Default: 0.
 
         Returns
         -------
@@ -261,15 +279,10 @@ class Filter:
             Average spectral luminosity (Lnu) or flux (Fnu) in the filter
         """
         freq *= (1. + z)
-        wl = (const.c / freq).to(u.angstrom).value
-        A = np.squeeze([extinction_law(wl, 3.1 * e) for e in np.atleast_1d(ebv)])
-
-        self.trans.sort('freq')
-        T_per_freq = self.trans['T'].quantity / self.trans['freq'].quantity
-        T_interp = np.interp(freq.value, self.trans['freq'].data, T_per_freq.value, left=0., right=0.)
+        T_per_freq = self.trans['T'].value / self.trans['freq'].value
+        T_interp = np.interp(freq, self.trans['freq'][::-1].value, T_per_freq[::-1], left=0., right=0.)
         T_norm_per_freq = T_interp / np.trapz(T_interp, freq)
-
-        return np.trapz(lum * 10. ** (A / -2.5) * T_norm_per_freq, freq)
+        return np.trapz(lum * extinction_law(freq, ebv) * T_norm_per_freq, freq)
 
     def __str__(self):
         return self.name
@@ -339,6 +352,24 @@ all_filters = [
     Filter('H', '#888888', -3, 'UKIRT', 1.021e-23, filename='Gemini_Flamingos2.H.dat', angstrom=True),
     Filter(['K', 'Ks'], '#CCCCCC', -4, 'UKIRT', 0.640e-23, filename='Gemini_Flamingos2.Ks.dat', angstrom=True),
     Filter('L', 'r', -4, 'UKIRT', 0.285e-23),
+    # JWST
+    Filter('F070W', 'C7', 0, 'JWST NIRCam', filename='JWST_NIRCam.F070W.dat', angstrom=True, italics=False),
+    Filter('F090W', 'C0', 0, 'JWST NIRCam', filename='JWST_NIRCam.F090W.dat', angstrom=True, italics=False),
+    Filter('F115W', 'C8', 0, 'JWST NIRCam', filename='JWST_NIRCam.F115W.dat', angstrom=True, italics=False),
+    Filter('F150W', 'C1', 0, 'JWST NIRCam', filename='JWST_NIRCam.F150W.dat', angstrom=True, italics=False),
+    Filter('F200W', 'C2', 0, 'JWST NIRCam', filename='JWST_NIRCam.F200W.dat', angstrom=True, italics=False),
+    Filter('F277W', 'C3', 0, 'JWST NIRCam', filename='JWST_NIRCam.F277W.dat', angstrom=True, italics=False),
+    Filter('F356W', 'C4', 0, 'JWST NIRCam', filename='JWST_NIRCam.F356W.dat', angstrom=True, italics=False),
+    Filter('F444W', 'C5', 0, 'JWST NIRCam', filename='JWST_NIRCam.F444W.dat', angstrom=True, italics=False),
+    Filter('F560W', 'C9', 0, 'JWST MIRI', filename='JWST_MIRI.F560W.dat', angstrom=True, italics=False),
+    Filter('F770W', 'C6', 0, 'JWST MIRI', filename='JWST_MIRI.F770W.dat', angstrom=True, italics=False),
+    Filter('F1000W', 'C7', 0, 'JWST MIRI', filename='JWST_MIRI.F1000W.dat', angstrom=True, italics=False),
+    Filter('F1130W', 'C0', 0, 'JWST MIRI', filename='JWST_MIRI.F1130W.dat', angstrom=True, italics=False),
+    Filter('F1280W', 'C8', 0, 'JWST MIRI', filename='JWST_MIRI.F1280W.dat', angstrom=True, italics=False),
+    Filter('F1500W', 'C1', 0, 'JWST MIRI', filename='JWST_MIRI.F1500W.dat', angstrom=True, italics=False),
+    Filter('F1800W', 'C9', 0, 'JWST MIRI', filename='JWST_MIRI.F1800W.dat', angstrom=True, italics=False),
+    Filter('F2100W', 'C2', 0, 'JWST MIRI', filename='JWST_MIRI.F2100W.dat', angstrom=True, italics=False),
+    Filter('F2550W', 'C3', 0, 'JWST MIRI', filename='JWST_MIRI.F2550W.dat', angstrom=True, italics=False),
     Filter(['unknown', '?'], 'w', 0, 'unknown', linecolor='k', italics=False)]
 Filter.order = [f.name for f in all_filters]
 filtdict = {}
