@@ -21,11 +21,12 @@ def lightcurve_mcmc(lc, model, priors=None, p_min=None, p_max=None, p_lo=None, p
     lc : lightcurve_fitting.lightcurve.LC
         Table of broadband photometry including columns "MJD", "mag", "dmag", "filt"
     model : lightcurve_fitting.models.Model
-        The model to fit to the light curve. Available models: :class:`models.ShockCooling`,
-        :class:`models.ShockCooling2`, :class:`models.CompanionShocking`
+        The model to fit to the light curve. Available models: :class:`.models.ShockCooling`,
+        :class:`.models.ShockCooling2`, :class:`.models.ShockCooling3`, :class:`.models.CompanionShocking`,
+        :class:`.models.CompanionShocking2`
     priors : list, optional
         Prior probability distributions for each model parameter. Available priors:
-        :func:`models.UniformPrior` (default), :func:`models.LogUniformPrior`, :func:`models.GaussianPrior`
+        :class:`.models.UniformPrior` (default), :class:`.models.LogUniformPrior`, :class:`.models.GaussianPrior`
     p_min, p_max : list, optional
         DEPRECATED: Use `priors` instead
     p_lo : list
@@ -115,6 +116,12 @@ def lightcurve_mcmc(lc, model, priors=None, p_min=None, p_max=None, p_lo=None, p
     elif len(priors) != ndim:
         raise Exception('priors must have length {:d}'.format(ndim))
 
+    for param, prior, p0, p1 in zip(model.input_names, priors, p_lo, p_up):
+        if p0 < prior.p_min:
+            raise Exception(f'starting guess for {param} (p_lo = {p0}) is outside prior (p_min = {prior.p_min})')
+        if p1 > prior.p_max:
+            raise Exception(f'starting guess for {param} (p_up = {p1}) is outside prior (p_max = {prior.p_max})')
+
     if sigma_type == 'relative':
         sigma_units = dy
     elif sigma_type == 'absolute':
@@ -374,7 +381,13 @@ def format_credible_interval(x, sigfigs=1, percentiles=(15.87, 50., 84.14), axis
     uncertainties = np.diff(quantiles)
     smaller_unc = np.amin(uncertainties, axis=-1)
     log_unc = np.log10(smaller_unc)
-    roundto = sigfigs - np.ceil(log_unc).astype(int)
+    roundto = sigfigs - np.floor(log_unc).astype(int) - 1
+
+    # catch numbers that will have more sigfigs after rounding
+    smaller_unc_round = [np.round(unc, dec) for unc, dec in zip(smaller_unc, roundto)]
+    log_unc_round = np.log10(smaller_unc_round)
+    roundto = sigfigs - np.floor(log_unc_round).astype(int) - 1
+
     quantiles = np.atleast_2d(quantiles)
     uncertainties = np.atleast_2d(uncertainties)
     roundto = np.atleast_1d(roundto)
