@@ -41,7 +41,7 @@ column_names = {
     'Apparent Magnitude': ['mag', 'Magnitude', 'Mag', 'ab_mag', 'PSFmag', 'MAG', 'omag', 'magnitude', 'apparent_mag'],
     'Apparent Magnitude Uncertainty': [
         'dmag', 'Magnitude_Error', 'magerr', 'MagErr', 'mag_err', 'e_mag', 'Error', 'err', 'PSFerr', 'MAGERR', 'e_omag',
-        'e_magnitude', 'apparent_mag_err', 'Mag_Err',
+        'e_magnitude', 'apparent_mag_err', 'Mag_Err', 'emag',
     ],
     'MJD': ['MJD', 'mjd'],
     'JD': ['JD', 'jd'],
@@ -52,12 +52,13 @@ column_names = {
     'Absolute Magnitude': ['absmag'],
     'Luminosity $L_ν$ (W Hz$^{-1}$)': ['lum'],
     'Luminosity Uncertainty': ['dlum'],
+    'Effective Wavelength (nm)': ['wl_eff'],  # calculated from filters; does not need to be in usage.rst
 }
 
 
 class LC(Table):
     """
-    A broadband light curve, stored as an :class:`astropy.table.Table` (see ``help(Table)`` for more details)
+    A broadband light curve, stored as an :class:`astropy.table.Table`
 
     Attributes
     ----------
@@ -341,10 +342,13 @@ class LC(Table):
         peaktable = self.where(**criteria)
         if len(peaktable):
             imin = np.argmin(peaktable['mag'])
-            self.sn.peakdate = peaktable['MJD'][imin]
+            self.meta['peakdate'] = peaktable['MJD'][imin]
         else:
-            self.sn.peakdate = np.nan
-        self.sn.peakcriteria = criteria
+            self.meta['peakdate'] = np.nan
+        if self.sn is not None:
+            self.meta['redshift'] = self.sn.z  # needed for calcPhase
+            self.sn.peakdate = self.meta['peakdate']
+            self.sn.peakcriteria = criteria
 
     def calcPhase(self, rdsp=False, hours=False):
         """
@@ -409,6 +413,10 @@ class LC(Table):
             Keyword arguments matching column names in the light curve are used to specify a subset of points to plot.
             Additional keyword arguments passed to :func:`matplotlib.pyplot.plot`.
         """
+        if xcol.startswith('filter'):
+            unit = xcol.split(':')[-1] if ':' in xcol else None
+            xcol = 'wl_eff'
+            self[xcol] = [f.wl_eff.to(unit) if unit else f.wl_eff for f in self['filter']]
         xchoices = ['phase', 'MJD']
         while xcol not in self.keys():
             xchoices.remove(xcol)
